classdef finiteElements2D < finiteElements
    % finiteElements
    % Abstract calls to build Finite Elements
    
    % Change log 
    % * June 2014:
    % Restruct the code for the matrix assembling. New method  
    % getConstantPartOfStiffnessMatrix introduced and signature of makeJ
    % We now compute matrices such that K = c(1)*K1+c(2)*K2 where c1/2 are
    % the vectors from the matrix [c1(x) 0    ]
    %                             [0     c2(x)] 
    
    methods(Access = public)        
        function [K,M,F] = createMatrixEntries(obj,gridObj,cf,af,ff)
            % the sizes of elementary matrices, varies from element
            % type to element type...
            % The space needed by sparse methos. Note that S1 etc are
            % stored as vectors. It length is 9 in the case of linear
            % Elements on triangles but it can vary 
            sizeVector = sqrt(size(obj.S1,1));
            sizeMatrix = size(obj.S1,1);
                        
            % compute the coefficients. The result is always a k x nElelemts
            % matrix where k = 3 for c, k = 1 for a and f.
            [cval,aval,fval] = obj.aCoefficients(gridObj,cf,af,ff);
            % We obtain the Matrices from the exacedt integration multiplied
            % by the grid-dependent but for fix grid constant
            % transformations.
            [K1,K2] = obj.getConstantPartOfStiffnessMatrix(gridObj);           
            % Now we have to multiply the Ks by c(1) and c(2).
            % For the mixed derivatives, this code must be EXTENDED.
            K = reshape(K1*sparse(1:gridObj.nElements,...
                1:gridObj.nElements,cval(1,:))...
                + K2*sparse(1:gridObj.nElements,1:...
                gridObj.nElements,cval(2,:))...
                    ,1,gridObj.nElements*sizeMatrix);
            
            M = reshape(obj.M*(aval.*obj.makeJ(gridObj)),1,...
                gridObj.nElements*sizeMatrix);
            
            F = reshape(obj.F*(obj.makeJ(gridObj).*fval),1,...
                gridObj.nElements*sizeVector);
        end
        
        function val = createConvectionEntries(obj,gridObj,b)
            % the point of the triagles
            p1 = gridObj.p(:,(gridObj.t(1,:)));
            p2 = gridObj.p(:,(gridObj.t(2,:)));
            p3 = gridObj.p(:,(gridObj.t(3,:)));
            % get element corner coordinates
            x21 = p2(1,:)-p1(1,:);
            x31 = p3(1,:)-p1(1,:);
            y21 = p2(2,:)-p1(2,:);
            y31 = p3(2,:)-p1(2,:);
            bval = obj.convCoefficients(gridObj,b);
            J1 = bval(:,1)'.*y31-bval(:,2)'.*x31; 
            J2 = bval(:,2)'.*x21-bval(:,1)'.*y21; 
            val = reshape(obj.C1*J1+obj.C2*J2,1,gridObj.nElements*9);        
        end          
    end
    
    methods(Access = public,Hidden)
        function [K1,K2] = getConstantPartOfStiffnessMatrix(obj,gridObj)
            % Needed by createMatrixEntries
            % helper method to compute the former abstract J 
            %                        
            p1 = gridObj.p(:,(gridObj.t(1,:)));
            p2 = gridObj.p(:,(gridObj.t(2,:)));
            p3 = gridObj.p(:,(gridObj.t(3,:))); 
            
            % get element corner coordinates
            x21 = p2(1,:)-p1(1,:);
            x31 = p3(1,:)-p1(1,:); 
            
            y21 = p2(2,:)-p1(2,:);
            y31 = p3(2,:)-p1(2,:); 
            
            % now we compute J for every element in one operation              
            J = obj.makeJ(gridObj) ;    
            % Inversion by using the famous 2x2 inversion formula, but with
            % out division by J: A few lines later we compute e.g.
            % xi_x^2*j so we can "correct" this here...
            xi_x  =  y31; 
            eta_x = -y21;
            xi_y = -x31;
            eta_y = x21;  
            
            K1 = obj.S1*((xi_x.^2)./J) + obj.S3*((eta_x.^2)./J)...
                +  obj.S2*((xi_x.*eta_x)./J);
            
            K2 = obj.S1*((xi_y.^2)./J) + obj.S3*((eta_y.^2)./J)...
                +  obj.S2*((xi_y.*eta_y)./J);
        end
    end
    
    methods(Static,Access=protected)
        function [cval,aval,fval] =  aCoefficients(gridObj,cc,aa,ff)
            [cval,aval,fval] = gridObj.aCoefficientsMpt(cc,aa,ff);
        end       
        
        function bval = convCoefficients(gridObj,b)
            bval = gridObj.convCoefficientsMpt(b);
        end
     
        
    end
    
    methods(Static,Access=public)
        function [Q,G,H,R] = assemb(gridObj)               
            if ~isa(gridObj,'gridd')
                throw(finiteElements.wrongClass);
            end   
            Q = spalloc( gridObj.nPoints, gridObj.nPoints,gridObj.nPoints);
            H = spalloc( gridObj.nPoints, gridObj.nPoints,gridObj.nPoints);
            G = spalloc( gridObj.nPoints,1, gridObj.nPoints); 
            R = spalloc( gridObj.nPoints,1,  gridObj.nPoints);
            
            
            p = gridObj.p;
            e = gridObj.e; 
            b = gridObj.b;
                
            
            for k = 1:gridObj.nEdges
                % length of the egde segment
                l = sqrt((p(1,e(2,k))-p(1,e(1,k)))^2 + ...
                    (p(2,e(2,k))-p(2,e(1,k)))^2);
                
                % simple L1 case, same as in assemb
                % first testing if coefficients are parametrized by
                % arclength
                if is_arclength_parametrized(b(:,e(5,k)))
                    midpoint_of_the_edge = 0.5*( e(3,k)+e(4,k));
                else
                    midpoint_of_the_edge = 0.5*[p(1,e(2,k))+p(1,e(1,k))...
                        ,p(2,e(2,k))+p(2,e(1,k))];
                end
                % where we put the entries...
                indx = e(1:2,k);
                
                [qval,gval,hval,rval] = gridObj.boundCoefficients(...
                    midpoint_of_the_edge,b(:,e(5,k)));
                
                % Here we postpone the "clever sparse action" and accept a
                % slower code...
                Q(indx, indx) = Q(indx, indx)+...
                    qval*l*reshape(lagrange11D.M,2,2); %#ok<*SPRIX>
                G(indx) = G(indx)+gval*l*lagrange11D.F; 
                H(indx, indx) = H(indx,indx)+hval*diag([1 1],0);
                R(indx) = R(indx)+rval*[1;1];     
            end
         
            
            
            
            function xy = is_arclength_parametrized(b)
                % checks if the function in b are arc-length-parametrized  
                % or not by trial-and-error.
                % Idee: If (x,y) is given and b is a function of b(x,y) 
                % or b(x) or b(y), 
                xy = false;  
                x = 0; y = 0; %#ok<*NASGU> 
                % We need x and y for evaluation of string expressions  
                % like 'sin(x)+y'
                if b(2) == 0, 
                    lengthq = b(3);
                    lengthg = b(4);
                    % Neumann  BC on this edge segment       
                    try % checks if g = g(x,y) or g = const                 
                        % if it works, q and g are written in terms of (x,y)
                        a1 = eval(char(b(5:5+b(3)-1)));
                        a2 = eval(char(b(5+lengthq:5+lengthq+lengthg-1)));
                        if ~(isnumeric(a1)&&isnumeric(a2))
                            throw(finiteElement.wrongClass)
                        end
                    catch ME
                        % Tricky: If an exception was thrown, it must be
                        % parametrisied by x an d y variables.
                        xy = true;                         
                    end
                else
                    % Dirichlet  BC on this edge segment
                    try
                        % if it works, h is written in terms of (x,y)
                        lengthh = b(5);
                        lengthr = b(6);
                        a1 = eval(char(b(9:9+lengthh-1)));
                        a2 = eval(char(b(9+lengthh:9+lengthh+lengthr-1)));
                        if ~(isnumeric(a1)&&isnumeric(a2))
                            throw(finiteElement.wrongClass)
                        end
                    catch ME 
                        xy = true; 
                    end
                end
            end
        end
        
        function J = makeJ(gridObj)   
            % compute Jacobi determinat
            p = gridObj.p;
            t = gridObj.t;
            p1 = p(:,(t(1,:)));
            p2 = p(:,(t(2,:)));
            p3 = p(:,(t(3,:)));
            x21 = p2(1,:)-p1(1,:);
            x31 = p3(1,:)-p1(1,:);
            y21 = p2(2,:)-p1(2,:);
            y31 = p3(2,:)-p1(2,:);            
            J = x21.*y31-x31.*y21;   
        end 
    end
end

